from __future__ import annotations

from typing import TYPE_CHECKING, Any

from daft.context import get_context
from daft.daft import (
    FileFormat,
    IOConfig,
    JoinType,
    PyExpr,
    PySchema,
    ResourceRequest,
    ScanTask,
    WriteMode,
)
from daft.execution import execution_step, physical_plan
from daft.expressions import Expression, ExpressionsProjection
from daft.logical.map_partition_ops import MapPartitionOp
from daft.logical.schema import Schema
from daft.runners.partitioning import PartitionT

if TYPE_CHECKING:
    from pyiceberg.schema import Schema as IcebergSchema
    from pyiceberg.table import TableProperties as IcebergTableProperties

    from daft.io import DataSink
    from daft.recordbatch import MicroPartition


def scan_with_tasks(
    scan_tasks: list[ScanTask],
) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
    """child_plan represents partitions with filenames.

    Yield a plan to read those filenames.
    """
    # TODO(Clark): Currently hardcoded to have 1 file per instruction
    # We can instead right-size and bundle the ScanTask into single-instruction bulk reads.

    cfg = get_context().daft_execution_config

    for scan_task in scan_tasks:
        scan_step = execution_step.PartitionTaskBuilder[PartitionT](
            inputs=[],
            partial_metadatas=None,
        ).add_instruction(
            instruction=execution_step.ScanWithTask(scan_task),
            # Set the estimated in-memory size as the memory request.
            resource_request=ResourceRequest(memory_bytes=scan_task.estimate_in_memory_size_bytes(cfg)),
        )
        yield scan_step


def empty_scan(
    schema: Schema,
) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
    """Yield a plan to create an empty Partition."""
    scan_step = execution_step.PartitionTaskBuilder[PartitionT](
        inputs=[],
        partial_metadatas=None,
    ).add_instruction(
        instruction=execution_step.EmptyScan(schema=schema),
        resource_request=ResourceRequest(memory_bytes=0),
    )
    yield scan_step


def project(
    input: physical_plan.InProgressPhysicalPlan[PartitionT],
    projection: list[PyExpr],
    resource_request: ResourceRequest | None,
) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
    expr_projection = ExpressionsProjection([Expression._from_pyexpr(expr) for expr in projection])
    resource_request = (
        ResourceRequest(num_cpus=1, num_gpus=None, memory_bytes=None) if resource_request is None else resource_request
    )
    return physical_plan.pipeline_instruction(
        child_plan=input,
        pipeable_instruction=execution_step.Project(expr_projection),
        resource_request=resource_request,
    )


def actor_pool_project(
    input: physical_plan.InProgressPhysicalPlan[PartitionT],
    projection: list[PyExpr],
    actor_pool_manager: physical_plan.ActorPoolManager,
    resource_request: ResourceRequest | None,
    num_actors: int,
) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
    expr_projection = ExpressionsProjection([Expression._from_pyexpr(expr) for expr in projection])
    resource_request = (
        ResourceRequest(num_cpus=1, num_gpus=None, memory_bytes=None) if resource_request is None else resource_request
    )
    return physical_plan.actor_pool_project(
        child_plan=input,
        projection=expr_projection,
        actor_pool_manager=actor_pool_manager,
        resource_request=resource_request,
        num_actors=num_actors,
    )


class ShimExplodeOp(MapPartitionOp):
    explode_columns: ExpressionsProjection

    def __init__(self, explode_columns: ExpressionsProjection) -> None:
        self.explode_columns = explode_columns

    def get_output_schema(self) -> Schema:
        raise NotImplementedError("Output schema shouldn't be needed at execution time")

    def run(self, input_partition: MicroPartition) -> MicroPartition:
        return input_partition.explode(self.explode_columns)


def explode(
    input: physical_plan.InProgressPhysicalPlan[PartitionT], explode_exprs: list[PyExpr]
) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
    explode_expr_projection = ExpressionsProjection(
        [Expression._from_pyexpr(expr)._explode() for expr in explode_exprs]
    )
    explode_op = ShimExplodeOp(explode_expr_projection)
    return physical_plan.pipeline_instruction(
        child_plan=input,
        pipeable_instruction=execution_step.MapPartition(explode_op),
        resource_request=ResourceRequest(),
    )


def unpivot(
    input: physical_plan.InProgressPhysicalPlan[PartitionT],
    ids: list[PyExpr],
    values: list[PyExpr],
    variable_name: str,
    value_name: str,
) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
    ids_projection = ExpressionsProjection([Expression._from_pyexpr(expr) for expr in ids])
    values_projection = ExpressionsProjection([Expression._from_pyexpr(expr) for expr in values])

    unpivot_step = execution_step.Unpivot(
        ids=ids_projection,
        values=values_projection,
        variable_name=variable_name,
        value_name=value_name,
    )

    return physical_plan.pipeline_instruction(
        child_plan=input,
        pipeable_instruction=unpivot_step,
        resource_request=ResourceRequest(),
    )


def local_aggregate(
    input: physical_plan.InProgressPhysicalPlan[PartitionT],
    agg_exprs: list[PyExpr],
    group_by: list[PyExpr],
) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
    aggregation_step = execution_step.Aggregate(
        to_agg=[Expression._from_pyexpr(pyexpr) for pyexpr in agg_exprs],
        group_by=ExpressionsProjection([Expression._from_pyexpr(pyexpr) for pyexpr in group_by]),
    )

    return physical_plan.pipeline_instruction(
        child_plan=input,
        pipeable_instruction=aggregation_step,
        resource_request=ResourceRequest(),
    )


def local_dedup(
    input: physical_plan.InProgressPhysicalPlan[PartitionT],
    columns: list[PyExpr],
) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
    dedup_step = execution_step.Dedup(
        columns=ExpressionsProjection([Expression._from_pyexpr(pyexpr) for pyexpr in columns]),
    )
    return physical_plan.pipeline_instruction(
        child_plan=input,
        pipeable_instruction=dedup_step,
        resource_request=ResourceRequest(),
    )


def pivot(
    input: physical_plan.InProgressPhysicalPlan[PartitionT],
    group_by: list[PyExpr],
    pivot_col: PyExpr,
    value_col: PyExpr,
    names: list[str],
) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
    pivot_step = execution_step.Pivot(
        group_by=ExpressionsProjection([Expression._from_pyexpr(pyexpr) for pyexpr in group_by]),
        pivot_col=Expression._from_pyexpr(pivot_col),
        value_col=Expression._from_pyexpr(value_col),
        names=names,
    )

    return physical_plan.pipeline_instruction(
        child_plan=input,
        pipeable_instruction=pivot_step,
        resource_request=ResourceRequest(),
    )


def sample(
    input: physical_plan.InProgressPhysicalPlan[PartitionT], fraction: float, with_replacement: bool, seed: int | None
) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
    return physical_plan.pipeline_instruction(
        child_plan=input,
        pipeable_instruction=execution_step.Sample(fraction=fraction, with_replacement=with_replacement, seed=seed),
        resource_request=ResourceRequest(),
    )


def sort(
    input: physical_plan.InProgressPhysicalPlan[PartitionT],
    sort_by: list[PyExpr],
    descending: list[bool],
    nulls_first: list[bool],
    num_partitions: int,
) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
    expr_projection = ExpressionsProjection([Expression._from_pyexpr(expr) for expr in sort_by])
    return physical_plan.sort(
        child_plan=input,
        sort_by=expr_projection,
        descending=descending,
        nulls_first=nulls_first,
        num_partitions=num_partitions,
    )


def top_n(
    input: physical_plan.InProgressPhysicalPlan[PartitionT],
    sort_by: list[PyExpr],
    descending: list[bool],
    nulls_first: list[bool],
    limit: int,
    num_partitions: int,
) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
    expr_projection = ExpressionsProjection([Expression._from_pyexpr(expr) for expr in sort_by])
    return physical_plan.top_n(
        child_plan=input,
        sort_by=expr_projection,
        descending=descending,
        nulls_first=nulls_first,
        limit=limit,
        num_partitions=num_partitions,
    )


def fanout_by_hash(
    input: physical_plan.InProgressPhysicalPlan[PartitionT],
    num_partitions: int,
    partition_by: list[PyExpr],
) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
    expr_projection = ExpressionsProjection([Expression._from_pyexpr(expr) for expr in partition_by])
    fanout_instruction = execution_step.FanoutHash(
        _num_outputs=num_partitions,
        partition_by=expr_projection,
    )
    return physical_plan.pipeline_instruction(
        input,
        fanout_instruction,
        ResourceRequest(),
    )


def reduce_merge(
    input: physical_plan.InProgressPhysicalPlan[PartitionT],
) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
    reduce_instruction = execution_step.ReduceMerge()
    return physical_plan.reduce(input, reduce_instruction)


def hash_join(
    input: physical_plan.InProgressPhysicalPlan[PartitionT],
    right: physical_plan.InProgressPhysicalPlan[PartitionT],
    left_on: list[PyExpr],
    right_on: list[PyExpr],
    null_equals_nulls: list[bool] | None,
    join_type: JoinType,
) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
    left_on_expr_proj = ExpressionsProjection([Expression._from_pyexpr(expr) for expr in left_on])
    right_on_expr_proj = ExpressionsProjection([Expression._from_pyexpr(expr) for expr in right_on])
    return physical_plan.hash_join(
        left_plan=input,
        right_plan=right,
        left_on=left_on_expr_proj,
        right_on=right_on_expr_proj,
        how=join_type,
        null_equals_nulls=null_equals_nulls,
    )


def merge_join_sorted(
    input: physical_plan.InProgressPhysicalPlan[PartitionT],
    right: physical_plan.InProgressPhysicalPlan[PartitionT],
    left_on: list[PyExpr],
    right_on: list[PyExpr],
    join_type: JoinType,
    left_is_larger: bool,
) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
    left_on_expr_proj = ExpressionsProjection([Expression._from_pyexpr(expr) for expr in left_on])
    right_on_expr_proj = ExpressionsProjection([Expression._from_pyexpr(expr) for expr in right_on])
    return physical_plan.merge_join_sorted(
        left_plan=input,
        right_plan=right,
        left_on=left_on_expr_proj,
        right_on=right_on_expr_proj,
        how=join_type,
        left_is_larger=left_is_larger,
    )


def sort_merge_join_aligned_boundaries(
    input: physical_plan.InProgressPhysicalPlan[PartitionT],
    right: physical_plan.InProgressPhysicalPlan[PartitionT],
    left_on: list[PyExpr],
    right_on: list[PyExpr],
    join_type: JoinType,
    num_partitions: int,
    left_is_larger: bool,
) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
    left_on_expr_proj = ExpressionsProjection([Expression._from_pyexpr(expr) for expr in left_on])
    right_on_expr_proj = ExpressionsProjection([Expression._from_pyexpr(expr) for expr in right_on])
    return physical_plan.sort_merge_join_aligned_boundaries(
        left_plan=input,
        right_plan=right,
        left_on=left_on_expr_proj,
        right_on=right_on_expr_proj,
        how=join_type,
        num_partitions=num_partitions,
        left_is_larger=left_is_larger,
    )


def broadcast_join(
    broadcaster: physical_plan.InProgressPhysicalPlan[PartitionT],
    receiver: physical_plan.InProgressPhysicalPlan[PartitionT],
    left_on: list[PyExpr],
    right_on: list[PyExpr],
    null_equals_nulls: list[bool] | None,
    join_type: JoinType,
    is_swapped: bool,
) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
    left_on_expr_proj = ExpressionsProjection([Expression._from_pyexpr(expr) for expr in left_on])
    right_on_expr_proj = ExpressionsProjection([Expression._from_pyexpr(expr) for expr in right_on])
    return physical_plan.broadcast_join(
        broadcaster_plan=broadcaster,
        receiver_plan=receiver,
        left_on=left_on_expr_proj,
        right_on=right_on_expr_proj,
        how=join_type,
        is_swapped=is_swapped,
        null_equals_nulls=null_equals_nulls,
    )


def write_file(
    input: physical_plan.InProgressPhysicalPlan[PartitionT],
    write_mode: WriteMode,
    file_format: FileFormat,
    schema: PySchema,
    root_dir: str,
    compression: str | None,
    partition_cols: list[PyExpr] | None,
    io_config: IOConfig | None,
) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
    if partition_cols is not None:
        expr_projection = ExpressionsProjection([Expression._from_pyexpr(expr) for expr in partition_cols])
    else:
        expr_projection = None
    return physical_plan.file_write(
        input,
        write_mode,
        file_format,
        Schema._from_pyschema(schema),
        root_dir,
        compression,
        expr_projection,
        io_config,
    )


def write_iceberg(
    input: physical_plan.InProgressPhysicalPlan[PartitionT],
    base_path: str,
    iceberg_schema: IcebergSchema,
    iceberg_properties: IcebergTableProperties,
    partition_spec_id: int,
    partition_cols: list[PyExpr],
    io_config: IOConfig | None,
) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
    return physical_plan.iceberg_write(
        input,
        base_path=base_path,
        iceberg_schema=iceberg_schema,
        iceberg_properties=iceberg_properties,
        partition_spec_id=partition_spec_id,
        partition_cols=ExpressionsProjection([Expression._from_pyexpr(expr) for expr in partition_cols]),
        io_config=io_config,
    )


def write_deltalake(
    input: physical_plan.InProgressPhysicalPlan[PartitionT],
    path: str,
    large_dtypes: bool,
    version: int,
    partition_cols: list[PyExpr] | None,
    io_config: IOConfig | None,
) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
    return physical_plan.deltalake_write(
        input,
        path,
        large_dtypes,
        version,
        ExpressionsProjection([Expression._from_pyexpr(expr) for expr in partition_cols]) if partition_cols else None,
        io_config,
    )


def write_lance(
    input: physical_plan.InProgressPhysicalPlan[PartitionT],
    path: str,
    mode: str,
    io_config: IOConfig | None,
    kwargs: dict[str, Any] | None,
) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
    return physical_plan.lance_write(input, path, mode, io_config, kwargs)


def write_data_sink(
    input: physical_plan.InProgressPhysicalPlan[PartitionT],
    sink: DataSink[Any],
) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
    return physical_plan.data_sink_write(input, sink)



from __future__ import annotations

import re
import string
import unicodedata

import pytest

from daft.expressions import col
from daft.recordbatch import MicroPartition


# source: RedPajama
def manual_normalize(text, remove_punct, lowercase, nfd_unicode, white_space):
    if text is None:
        return None

    if remove_punct:
        text = text.translate(str.maketrans("", "", string.punctuation))

    if lowercase:
        text = text.lower()

    if white_space:
        text = text.strip()
        text = re.sub(r"\s+", " ", text)

    if nfd_unicode:
        text = unicodedata.normalize("NFD", text)

    return text


NORMALIZE_TEST_DATA = [
    "regular text no changes",
    "Regular texT WITH uPpErCaSe",
    "ALL UPPERCASE TEXT",
    "text, with... punctuation!!!",
    "!&# #%*!()%*@# &*%#& @*( #*(@%()))",
    "!@#$%^&*()+_-=~`[]\\/.,';?><\":|}{",
    "UPPERCASE, AND, PUNCTUATION!?",
    "füñķÿ úňìčõðė",
    "füñķÿ, úňìčõðė!",
    "FüÑķŸ úňÌčõÐė",
    "FüÑķŸ,   úňÌčõÐė!",
    "way    too    much      space",
    "     space  all     over   the place    ",
    "other\ntypes\tof\r\nspace characters",
    "too\n\n\t\r\n  \n\t\tmuch\n\t\tspace\t  \n\n  \t\r\n \t",
    None,
    "TOO\n\n\t\r\n  \n\t\tMUCH\n\t\tsPACe\t  \n\n  \t\r\n \t",
    "too,\n\n?\t\r\n  \n\t\tmuc%h!!%\n\t\t\\SPACE??!\t  \n\n  \t\r\n \t",
    "FüÑķŸ,   úňÌčõÐė!   AND EVERYTHING else TOO    \t\na\t\t\nbCDe 😃😌😝",
    "",
    "specialcase",
    "SPECIALCASE",
    "😃",
    None,
]


@pytest.mark.parametrize("remove_punct", [False, True])
@pytest.mark.parametrize("lowercase", [False, True])
@pytest.mark.parametrize("nfd_unicode", [False, True])
@pytest.mark.parametrize("white_space", [False, True])
def test_utf8_normalize(remove_punct, lowercase, nfd_unicode, white_space):
    table = MicroPartition.from_pydict({"col": NORMALIZE_TEST_DATA})
    result = table.eval_expression_list(
        [
            col("col").str.normalize(
                remove_punct=remove_punct,
                lowercase=lowercase,
                nfd_unicode=nfd_unicode,
                white_space=white_space,
            )
        ]
    )
    expected = [manual_normalize(t, remove_punct, lowercase, nfd_unicode, white_space) for t in NORMALIZE_TEST_DATA]
    assert result.to_pydict() == {"col": expected}



from __future__ import annotations

import json

import pytest

import daft


def test_read_json_array():
    df = daft.read_json("tests/assets/json-data/sample1.json")
    with open("tests/assets/json-data/sample1.json") as f:
        expected = json.load(f)

    assert df.to_pylist() == expected


def test_read_json_multi():
    df = daft.read_json(["tests/assets/json-data/sample1.json", "tests/assets/json-data/sample2.json"])

    assert df.count_rows() == 25


def test_read_json_schema_is_correct():
    df = daft.read_json("tests/assets/json-data/sample1.json")
    expected_schema = daft.Schema.from_pydict(
        {"x": daft.DataType.int64(), "y": daft.DataType.string(), "z": daft.DataType.bool()}
    )

    assert df.schema() == expected_schema


def test_read_empty():
    with pytest.raises(Exception, match="Empty JSON file"):
        daft.read_json("tests/assets/json-data/empty")


def test_read_non_array():
    with pytest.raises(Exception, match="Invalid JSON format"):
        daft.read_json("tests/assets/json-data/baddata1.json")



from __future__ import annotations

import os
import sys
from unittest.mock import patch

import pytest

from daft.io.delta_lake.utils import construct_delta_file_path


@pytest.mark.skipif(sys.platform.startswith("win"), reason="Unix-specific test")
@pytest.mark.parametrize(
    "scheme,table_uri,relative_path,expected",
    [
        # Object storage schemes (unaffected by OS)
        ("s3", "s3://bucket/table", "data/file.parquet", "s3://bucket/table/data/file.parquet"),
        ("s3a", "s3a://bucket/table/", "data/file.parquet", "s3a://bucket/table/data/file.parquet"),
        # Local file system - Unix style
        ("file", "/tmp/table", "data/file.parquet", os.path.join("/tmp/table", "data/file.parquet")),
        (
            "hdfs",
            "hdfs://namenode/table",
            "data/file.parquet",
            os.path.join("hdfs://namenode/table", "data/file.parquet"),
        ),
    ],
)
def test_construct_delta_file_path(scheme, table_uri, relative_path, expected):
    result = construct_delta_file_path(scheme, table_uri, relative_path)
    assert result == expected


@pytest.mark.parametrize(
    "scheme,table_uri,relative_path,expected",
    [
        # Object storage schemes should always use forward slashes
        ("s3", "s3://bucket/table", "data/file.parquet", "s3://bucket/table/data/file.parquet"),
        ("s3a", "s3a://bucket/table/", "data/file.parquet", "s3a://bucket/table/data/file.parquet"),
        ("gcs", "gs://bucket/table", "_delta_log/00000.json", "gs://bucket/table/_delta_log/00000.json"),
        ("gs", "gs://bucket/table", "data/file.parquet", "gs://bucket/table/data/file.parquet"),
        ("az", "az://container/table", "data/file.parquet", "az://container/table/data/file.parquet"),
        ("abfs", "abfs://container/table", "data/file.parquet", "abfs://container/table/data/file.parquet"),
        ("abfss", "abfss://container/table", "data/file.parquet", "abfss://container/table/data/file.parquet"),
    ],
)
def test_object_storage_uses_forward_slashes_on_windows(scheme, table_uri, relative_path, expected):
    # Mock os.path.join to return Windows-style paths with backslashes
    with patch("os.path.join", return_value="C:\\fake\\windows\\path"):
        result = construct_delta_file_path(scheme, table_uri, relative_path)
        assert result == expected
        assert "\\" not in result  # Ensure no backslashes in object storage paths



from __future__ import annotations

import daft


def test_summarize_sanity():
    df = daft.from_pydict(
        {
            "A": [1, 2, 3, 4, 5],
            "B": [1.5, 2.5, 3.5, 4.5, 5.5],
            "C": [True, True, False, False, None],
            "D": [None, None, None, None, None],
        }
    )
    # row for each column of input
    assert df.summarize().count_rows() == 4
    assert df.select("A", "B").summarize().count_rows() == 2


def test_summarize_dataframe(make_df, valid_data: list[dict[str, float]]) -> None:
    df = daft.from_pydict(
        {
            "a": [1, 2, 3, 3],
            "b": [None, "a", "b", "c"],
        }
    )
    expected = {
        "column": ["a", "b"],
        "type": ["Int64", "Utf8"],
        "min": ["1", "a"],
        "max": ["3", "c"],
        "count": [4, 3],
        "count_nulls": [0, 1],
        "approx_count_distinct": [3, 3],
    }
    assert df.summarize().to_pydict() == expected



from __future__ import annotations

import os
from typing import Annotated

from fastapi import FastAPI, Header, Request, Response

from ..utils.parquet_generation import generate_parquet_file
from ..utils.request_range import parse_range_from_header
from ..utils.responses import get_response

BUCKET_NAME = "retries-parquet-bucket"
OBJECT_KEY_URL = "/{status_code}/{status_code_str}/{num_errors}/{item_id}"
MOCK_PARQUET_DATA_PATH = generate_parquet_file()

HEAD_NUM_RETRIES: dict[str, int] = {}
GET_NUM_RETRIES: dict[tuple[str, tuple[int, int]], int] = {}
NON_RETRYABLE_ERRORS = [400, 403, 404]

route = f"/{BUCKET_NAME}"
app = FastAPI()


@app.head(OBJECT_KEY_URL)
async def bucket_head(request: Request, status_code: int, status_code_str: str, num_errors: int, item_id: str):
    if status_code in NON_RETRYABLE_ERRORS:
        return get_response(request.url, status_code, status_code_str)

    key = item_id
    if key not in HEAD_NUM_RETRIES:
        HEAD_NUM_RETRIES[key] = 1
    else:
        HEAD_NUM_RETRIES[key] += 1
    if HEAD_NUM_RETRIES[key] <= num_errors:
        return get_response(request.url, status_code, status_code_str)

    return Response(
        headers={
            "Content-Length": str(os.path.getsize(MOCK_PARQUET_DATA_PATH.name)),
            "Content-Type": "binary/octet-stream",
            "Accept-Ranges": "bytes",
        },
    )


@app.get(OBJECT_KEY_URL)
async def bucket_get(
    request: Request,
    status_code: int,
    status_code_str: str,
    num_errors: int,
    item_id: str,
    range: Annotated[str, Header()],
):
    if status_code in NON_RETRYABLE_ERRORS:
        return get_response(request.url, status_code, status_code_str)

    # If we've only seen this range request <= num_errors times, we throw an error
    start, end = parse_range_from_header(range, os.path.getsize(MOCK_PARQUET_DATA_PATH.name))

    key = (item_id, (start, end))
    if key not in GET_NUM_RETRIES:
        GET_NUM_RETRIES[key] = 1
    else:
        GET_NUM_RETRIES[key] += 1
    if GET_NUM_RETRIES[key] <= num_errors:
        return get_response(request.url, status_code, status_code_str)

    with open(MOCK_PARQUET_DATA_PATH.name, "rb") as f:
        f.seek(start)
        data = f.read(end - start + 1)

    return Response(
        status_code=206,
        content=data,
        headers={
            "Content-Length": str(len(data)),
            "Content-Type": "binary/octet-stream",
            "Content-Range": f"bytes {start}-{end}/{os.path.getsize(MOCK_PARQUET_DATA_PATH.name)}",
        },
    )



from __future__ import annotations

import logging
import math
import warnings
from enum import Enum
from typing import TYPE_CHECKING, Any

from daft.context import get_context
from daft.daft import (
    DatabaseSourceConfig,
    FileFormatConfig,
    PyPartitionField,
    PyPushdowns,
    PyRecordBatch,
    ScanTask,
    StorageConfig,
)
from daft.expressions.expressions import lit
from daft.io.common import _get_schema_from_dict
from daft.io.scan import ScanOperator
from daft.recordbatch import RecordBatch

if TYPE_CHECKING:
    from collections.abc import Iterator

    from daft.datatype import DataType
    from daft.logical.schema import Schema
    from daft.sql.sql_connection import SQLConnection

logger = logging.getLogger(__name__)


class PartitionBoundStrategy(Enum):
    PERCENTILE = "percentile"
    MIN_MAX = "min-max"

    @classmethod
    def from_str(cls, value: str) -> PartitionBoundStrategy:
        try:
            return cls(value.lower())
        except ValueError:
            raise ValueError(f"Invalid PartitionBoundStrategy: {value}, must be either 'percentile' or 'min-max'")


class SQLScanOperator(ScanOperator):
    def __init__(
        self,
        sql: str,
        conn: SQLConnection,
        storage_config: StorageConfig,
        disable_pushdowns_to_sql: bool,
        infer_schema: bool,
        infer_schema_length: int,
        schema: dict[str, DataType] | None,
        partition_col: str | None = None,
        num_partitions: int | None = None,
        partition_bound_strategy: PartitionBoundStrategy | None = None,
    ) -> None:
        super().__init__()
        self.sql = sql
        self.conn = conn
        self.storage_config = storage_config
        self._disable_pushdowns_to_sql = disable_pushdowns_to_sql
        self._partition_col = partition_col
        self._num_partitions = num_partitions
        self._partition_bound_strategy = partition_bound_strategy
        self._schema = self._attempt_schema_read(infer_schema, infer_schema_length, schema)

    def schema(self) -> Schema:
        return self._schema

    def name(self) -> str:
        return "SQLScanOperator"

    def display_name(self) -> str:
        return f"SQLScanOperator(sql={self.sql}, conn={self.conn})"

    def partitioning_keys(self) -> list[PyPartitionField]:
        return []

    def multiline_display(self) -> list[str]:
        return [
            self.display_name(),
            f"Schema = {self._schema}",
        ]

    def to_scan_tasks(self, pushdowns: PyPushdowns) -> Iterator[ScanTask]:
        total_rows, total_size, num_scan_tasks = self._get_size_estimates()
        if num_scan_tasks == 0:
            return iter(())
        if num_scan_tasks == 1 or self._partition_col is None:
            return self._single_scan_task(pushdowns, total_rows, total_size)

        partition_bounds = self._get_partition_bounds(num_scan_tasks)
        partition_bounds_sql = [lit(bound)._to_sql() for bound in partition_bounds]

        if any(bound is None for bound in partition_bounds_sql):
            warnings.warn(
                "Unable to partition the data using the specified column. Falling back to a single scan task."
            )
            return self._single_scan_task(pushdowns, total_rows, total_size)

        size_bytes = (
            math.ceil(total_size / num_scan_tasks)
            if self._partition_bound_strategy == PartitionBoundStrategy.PERCENTILE
            else None
        )
        scan_tasks = []
        for i in range(num_scan_tasks):
            left_clause = f"{self._partition_col} >= {partition_bounds_sql[i]}"
            right_clause = (
                f"{self._partition_col} {'<' if i < num_scan_tasks - 1 else '<='} {partition_bounds_sql[i + 1]}"
            )
            stats = RecordBatch.from_pydict({self._partition_col: [partition_bounds[i], partition_bounds[i + 1]]})
            scan_task = self._construct_scan_task(
                pushdowns,
                num_rows=None,
                size_bytes=size_bytes,
                partition_bounds=(left_clause, right_clause),
                stats=stats._recordbatch,
            )
            scan_tasks.append(scan_task)

        return iter(scan_tasks)

    def can_absorb_filter(self) -> bool:
        return False

    def can_absorb_limit(self) -> bool:
        return False

    def can_absorb_select(self) -> bool:
        return False

    def _attempt_schema_read(
        self,
        infer_schema: bool,
        infer_schema_length: int,
        schema: dict[str, DataType] | None,
    ) -> Schema:
        # If schema is provided and user turned off schema inference, use the provided schema
        if schema is not None and not infer_schema:
            return _get_schema_from_dict(schema)

        # Else, attempt schema inference then apply the schema hint if provided
        inferred_schema = self.conn.read_schema(self.sql, infer_schema_length)
        if schema is not None:
            return inferred_schema.apply_hints(_get_schema_from_dict(schema))
        return inferred_schema

    def _get_size_estimates(self) -> tuple[int, float, int]:
        total_rows = self._get_num_rows()
        estimate_row_size_bytes = self.schema().estimate_row_size_bytes()
        total_size = total_rows * estimate_row_size_bytes
        num_scan_tasks = (
            math.ceil(total_size / get_context().daft_execution_config.read_sql_partition_size_bytes)
            if self._num_partitions is None
            else self._num_partitions
        )
        num_scan_tasks = min(num_scan_tasks, total_rows)
        return total_rows, total_size, num_scan_tasks

    def _get_num_rows(self) -> int:
        num_rows_sql = self.conn.construct_sql_query(self.sql, projection=["COUNT(*)"])
        pa_table = self.conn.execute_sql_query(num_rows_sql)

        if pa_table.num_rows != 1:
            raise RuntimeError(
                "Failed to get the number of rows: COUNT(*) query returned an unexpected number of rows."
            )
        if pa_table.num_columns != 1:
            raise RuntimeError(
                "Failed to get the number of rows: COUNT(*) query returned an unexpected number of columns."
            )

        return pa_table.column(0)[0].as_py()

    def _get_partition_bounds(self, num_scan_tasks: int) -> list[Any]:
        if self._partition_col is None:
            raise ValueError("Failed to get partition bounds: partition_col must be specified to partition the data.")

        if not (
            self._schema[self._partition_col].dtype.is_temporal()
            or self._schema[self._partition_col].dtype.is_numeric()
        ):
            raise ValueError(
                f"Failed to get partition bounds: {self._partition_col} is not a numeric or temporal type, and cannot be used for partitioning."
            )

        if self._partition_bound_strategy == PartitionBoundStrategy.PERCENTILE:
            try:
                # Try to get percentiles using percentile_disc.
                # Favor percentile_disc over percentile_cont because we want exact values to do <= and >= comparisons.
                percentiles = [i / num_scan_tasks for i in range(num_scan_tasks + 1)]
                # Use the OVER clause for SQL Server dialects
                over_clause = "OVER ()" if self.conn.dialect in ["mssql", "tsql"] else ""
                percentile_sql = self.conn.construct_sql_query(
                    self.sql,
                    projection=[
                        f"percentile_disc({percentile}) WITHIN GROUP (ORDER BY {self._partition_col}) {over_clause} AS bound_{i}"
                        for i, percentile in enumerate(percentiles)
                    ],
                    limit=1,
                )
                pa_table = self.conn.execute_sql_query(percentile_sql)

                if pa_table.num_rows != 1:
                    raise RuntimeError(f"Expected 1 row, but got {pa_table.num_rows}.")

                if pa_table.num_columns != num_scan_tasks + 1:
                    raise RuntimeError(f"Expected {num_scan_tasks + 1} percentiles, but got {pa_table.num_columns}.")

                pydict = RecordBatch.from_arrow_table(pa_table).to_pydict()
                assert pydict.keys() == {f"bound_{i}" for i in range(num_scan_tasks + 1)}
                return [pydict[f"bound_{i}"][0] for i in range(num_scan_tasks + 1)]

            except Exception as e:
                warnings.warn(
                    f"Failed to calculate partition bounds for read_sql using percentile strategy: {e!s}. "
                    "Falling back to MIN_MAX strategy."
                )
                self._partition_bound_strategy = PartitionBoundStrategy.MIN_MAX

        # Either MIN_MAX was explicitly specified or percentile calculation failed
        min_max_sql = self.conn.construct_sql_query(
            self.sql, projection=[f"MIN({self._partition_col}) as min", f"MAX({self._partition_col}) as max"]
        )
        pa_table = self.conn.execute_sql_query(min_max_sql)

        if pa_table.num_rows != 1:
            raise RuntimeError(f"Failed to get partition bounds: expected 1 row, but got {pa_table.num_rows}.")
        if pa_table.num_columns != 2:
            raise RuntimeError(f"Failed to get partition bounds: expected 2 columns, but got {pa_table.num_columns}.")

        pydict = RecordBatch.from_arrow_table(pa_table).to_pydict()
        assert pydict.keys() == {"min", "max"}
        min_val = pydict["min"][0]
        max_val = pydict["max"][0]
        range_size = (max_val - min_val) / num_scan_tasks
        return [min_val + range_size * i for i in range(num_scan_tasks)] + [max_val]

    def _single_scan_task(
        self, pushdowns: PyPushdowns, total_rows: int | None, total_size: float
    ) -> Iterator[ScanTask]:
        return iter([self._construct_scan_task(pushdowns, num_rows=total_rows, size_bytes=math.ceil(total_size))])

    def _construct_scan_task(
        self,
        pushdowns: PyPushdowns,
        num_rows: int | None = None,
        size_bytes: int | None = None,
        partition_bounds: tuple[str, str] | None = None,
        stats: PyRecordBatch | None = None,
    ) -> ScanTask:
        predicate_sql = pushdowns.filters.to_sql() if pushdowns.filters is not None else None
        apply_pushdowns_to_sql = not self._disable_pushdowns_to_sql and (
            pushdowns.filters is None or predicate_sql is not None
        )

        if apply_pushdowns_to_sql:
            sql = self.conn.construct_sql_query(
                self.sql,
                projection=pushdowns.columns,
                predicate=predicate_sql,
                limit=pushdowns.limit,
                partition_bounds=partition_bounds,
            )
        else:
            sql = self.conn.construct_sql_query(self.sql, partition_bounds=partition_bounds)

        file_format_config = FileFormatConfig.from_database_config(DatabaseSourceConfig(sql, self.conn))

        # keep column pushdowns because they are used for deriving the materialized schema
        remaining_pushdowns = pushdowns if not apply_pushdowns_to_sql else PyPushdowns(columns=pushdowns.columns)

        return ScanTask.sql_scan_task(
            url=self.conn.url,
            file_format=file_format_config,
            schema=self._schema._schema,
            storage_config=self.storage_config,
            num_rows=num_rows,
            size_bytes=size_bytes,
            pushdowns=remaining_pushdowns,
            stats=stats,
        )



from __future__ import annotations

import warnings
from urllib import request
from urllib.error import URLError
import uuid
import json
import os
from pathlib import Path
from daft.daft import dashboard as native
from importlib import resources
import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from datetime import datetime


def _should_run() -> bool:
    enable_dashboard_str = os.environ.get(native.DAFT_DASHBOARD_ENV_ENABLED)
    if not enable_dashboard_str:
        return False
    try:
        enable_dashboard = int(enable_dashboard_str)
    except ValueError:
        return False
    if not enable_dashboard:
        return False

    return True


def launch(noop_if_initialized: bool = False) -> None:
    """Launches the Daft dashboard server on port 3238.

    The server serves HTML/CSS/JS bundles, so you are able to point your browser towards `http://localhost:3238` and view information regarding your queries.

    # Arguments:
        - noop_if_initialized: bool = False
            Will not raise an exception a Daft dashboard server process is already launched and running.
            Otherwise, an exception will be raised.
    """
    os.environ[native.DAFT_DASHBOARD_ENV_ENABLED] = "1"

    handle = native.launch(noop_if_initialized=noop_if_initialized)

    import atexit

    atexit.register(handle.shutdown, noop_if_shutdown=True)


def broadcast_query_information(
    unoptimized_plan: str,
    optimized_plan: str,
    plan_time_start: datetime,
    plan_time_end: datetime,
) -> None:
    import ssl

    headers = {
        "Content-Type": "application/json",
    }

    if (url := os.environ.get("DAFT_DASHBOARD_URL")) is not None:
        url = f"{url}/queries"
    else:
        url = native.DAFT_DASHBOARD_QUERIES_URL

    if (auth_token := os.environ.get("DAFT_DASHBOARD_AUTH_TOKEN")) is not None:
        headers["Authorization"] = f"Bearer {auth_token}"

    data = json.dumps(
        {
            "id": str(uuid.uuid4()),
            "optimized_plan": optimized_plan,
            "run_id": os.environ.get("DAFT_DASHBOARD_RUN_ID", None),
            "logs": "",  # todo: implement logs
        }
    ).encode("utf-8")

    ctx = ssl.create_default_context()

    # if it's a localhost uri we can skip ssl verification
    if "localhost" in url:
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE

    req = request.Request(url, headers=headers, data=data)

    try:
        request.urlopen(req, timeout=1, context=ctx)
    except URLError as e:
        warnings.warn(f"Failed to broadcast metrics over {url}: {e}")


__all__ = [
    "_should_run",
    "broadcast_query_information",
    "launch",
]



from __future__ import annotations

from .expressions import (
    Expression,
    ExpressionsProjection,
    col,
    element,
    list_,
    lit,
    interval,
    struct,
    coalesce,
)

from .visitor import ExpressionVisitor

__all__ = [
    "Expression",
    "ExpressionVisitor",
    "ExpressionsProjection",
    "coalesce",
    "col",
    "element",
    "interval",
    "list_",
    "lit",
    "struct",
]



from __future__ import annotations

import itertools
from operator import and_, or_, xor

import pytest

from daft.datatype import DataType
from daft.series import Series

BITWISE_OPERATORS = [and_, or_, xor]
INT_TYPES = [
    DataType.int8(),
    DataType.int16(),
    DataType.int32(),
    DataType.int64(),
]
UINT_TYPES = [
    DataType.uint8(),
    DataType.uint16(),
    DataType.uint32(),
    DataType.uint64(),
]


@pytest.mark.parametrize("op", BITWISE_OPERATORS)
@pytest.mark.parametrize(
    # Test all combinations of signed and unsigned integers EXCEPT for UINT64 because UINT64 and INT types are not compatible with our supertype rules
    "dtype",
    itertools.product(INT_TYPES + UINT_TYPES[:-1], repeat=2),
)
def test_bitwise_with_int_and_unsigned_int_types(op, dtype):
    left_dtype, right_dtype = dtype
    s1 = Series.from_pylist([0b1100, 0b1010, 0b1001]).cast(left_dtype)
    s2 = Series.from_pylist([0b1010, 0b1100, 0b1001]).cast(right_dtype)

    result = op(s1, s2)
    expected = [op(i, j) for i, j in zip(s1.to_pylist(), s2.to_pylist())]
    assert result.to_pylist() == expected


@pytest.mark.parametrize("op", BITWISE_OPERATORS)
@pytest.mark.parametrize("dtype", itertools.product(UINT_TYPES, repeat=2))
def test_bitwise_with_different_unsigned_int_types(op, dtype):
    left_dtype, right_dtype = dtype
    s1 = Series.from_pylist([0b1100, 0b1010, 0b1001]).cast(left_dtype)
    s2 = Series.from_pylist([0b1010, 0b1100, 0b1001]).cast(right_dtype)

    result = op(s1, s2)
    expected = [op(i, j) for i, j in zip(s1.to_pylist(), s2.to_pylist())]
    assert result.to_pylist() == expected


@pytest.mark.parametrize(
    "left, right",
    [
        pytest.param([0b1100, 0b1010, 0b1001], [0b1010, 0b1100, 0b1001], id="same_length"),
        pytest.param([0b1100, 0b1010, 0b1001], [0b1010], id="broadcast_right"),
        pytest.param([0b1100], [0b1010, 0b1100, 0b1001], id="broadcast_left"),
    ],
)
@pytest.mark.parametrize("op", BITWISE_OPERATORS)
def test_bitwise_broadcasting(left, right, op):
    s1 = Series.from_pylist(left)
    s2 = Series.from_pylist(right)

    if len(left) == 1:
        left = left * len(right)
    if len(right) == 1:
        right = right * len(left)

    result = op(s1, s2)
    expected = [op(i, j) for i, j in zip(left, right)]
    assert result.to_pylist() == expected


@pytest.mark.parametrize(
    "left, right",
    [
        pytest.param([0b1100, None, 0b1001], [0b1010, 0b1100, None], id="same_length"),
        pytest.param([0b1100, 0b1010, 0b1001], [None], id="broadcast_right"),
        pytest.param([None], [0b1010, 0b1100, 0b1001], id="broadcast_left"),
    ],
)
@pytest.mark.parametrize("op", BITWISE_OPERATORS)
def test_bitwise_broadcasting_with_nulls(left, right, op):
    s1 = Series.from_pylist(left).cast(DataType.int8())
    s2 = Series.from_pylist(right).cast(DataType.int8())

    if len(left) == 1:
        left = left * len(right)
    if len(right) == 1:
        right = right * len(left)

    result = op(s1, s2)
    expected = [op(i, j) if i is not None and j is not None else None for i, j in zip(left, right)]
    assert result.to_pylist() == expected


@pytest.mark.parametrize("op", BITWISE_OPERATORS)
def test_bitwise_errors_with_non_integer_inputs(op):
    s1 = Series.from_pylist([0b1100, 0b1010, 0b1001])
    s2 = Series.from_pylist(["a", "b", "c"])

    with pytest.raises(ValueError, match="Cannot perform logic on types:"):
        op(s1, s2)


@pytest.mark.parametrize("op", BITWISE_OPERATORS)
def test_bitwise_errors_with_incompatible_lengths(op):
    s1 = Series.from_pylist([0b1100, 0b1010, 0b1001])
    s2 = Series.from_pylist([0b1010, 0b1100])

    with pytest.raises(ValueError, match="trying to operate on different length arrays:"):
        op(s1, s2)



from __future__ import annotations

import datetime
import uuid
import warnings
from typing import TYPE_CHECKING, Any

from daft import Expression, col, lit
from daft.datatype import DataType
from daft.expressions.expressions import ExpressionsProjection
from daft.io.common import _get_schema_from_dict
from daft.recordbatch import MicroPartition
from daft.recordbatch.partitioning import PartitionedTable, partition_strings_to_path

if TYPE_CHECKING:
    from collections.abc import Iterator

    from pyiceberg.manifest import DataFile
    from pyiceberg.partitioning import PartitionField as IcebergPartitionField
    from pyiceberg.schema import Schema as IcebergSchema
    from pyiceberg.table import TableProperties as IcebergTableProperties
    from pyiceberg.typedef import Record as IcebergRecord

    from daft.dependencies import pa, pads, pq


def get_missing_columns(data_schema: pa.Schema, iceberg_schema: IcebergSchema) -> ExpressionsProjection:
    """Add null values for columns in the schema that are missing from the table."""
    from pyiceberg.io.pyarrow import schema_to_pyarrow

    iceberg_pyarrow_schema = schema_to_pyarrow(iceberg_schema)

    existing_columns = set(data_schema.names)

    to_add = []
    for name in iceberg_pyarrow_schema.names:
        if name not in existing_columns:
            to_add.append(lit(None).alias(name).cast(DataType.from_arrow_type(iceberg_pyarrow_schema.field(name).type)))

    return ExpressionsProjection(to_add)


def coerce_pyarrow_table_to_schema(pa_table: pa.Table, schema: pa.Schema) -> pa.Table:
    """Coerces a PyArrow table to the supplied schema.

    1. For each field in `pa_table`, cast it to the field in `input_schema` if one with a matching name
        is available
    2. Reorder the fields in the casted table to the supplied schema, dropping any fields in `pa_table`
        that do not exist in the supplied schema
    3. If any fields in the supplied schema are not present, add a null array of the correct type

    This ensures that we populate field_id for iceberg as well as fill in null values where needed
    This might break for nested fields with large_strings
    we should test that behavior

    Args:
        pa_table (pa.Table): Table to coerce
        schema (pa.Schema): Iceberg schema to coerce to

    Returns:
        pa.Table: Table with schema == `schema`
    """
    import pyarrow as pa

    input_schema_names = set(schema.names)

    # Perform casting of types to provided schema's types
    cast_to_schema = [
        (schema.field(inferred_field.name) if inferred_field.name in input_schema_names else inferred_field)
        for inferred_field in pa_table.schema
    ]
    casted_table = pa_table.cast(pa.schema(cast_to_schema))

    # Reorder and pad columns with a null column where necessary
    pa_table_column_names = set(casted_table.column_names)
    columns = []
    for name in schema.names:
        if name in pa_table_column_names:
            columns.append(casted_table[name])
        else:
            columns.append(pa.nulls(len(casted_table), type=schema.field(name).type))
    return pa.table(columns, schema=schema)


def partition_field_to_expr(field: IcebergPartitionField, schema: IcebergSchema) -> Expression:
    from pyiceberg.transforms import (
        BucketTransform,
        DayTransform,
        HourTransform,
        IdentityTransform,
        MonthTransform,
        TruncateTransform,
        YearTransform,
    )

    part_col: Expression = col(schema.find_field(field.source_id).name)

    if isinstance(field.transform, IdentityTransform):
        transform_expr = part_col
    elif isinstance(field.transform, YearTransform):
        transform_expr = part_col.partitioning.years()
    elif isinstance(field.transform, MonthTransform):
        transform_expr = part_col.partitioning.months()
    elif isinstance(field.transform, DayTransform):
        transform_expr = part_col.partitioning.days()
    elif isinstance(field.transform, HourTransform):
        transform_expr = part_col.partitioning.hours()
    elif isinstance(field.transform, BucketTransform):
        transform_expr = part_col.partitioning.iceberg_bucket(field.transform.num_buckets)
    elif isinstance(field.transform, TruncateTransform):
        transform_expr = part_col.partitioning.iceberg_truncate(field.transform.width)
    else:
        warnings.warn(f"{field.transform} not implemented, Please make an issue!")
        transform_expr = part_col

    transform_expr = transform_expr.alias(field.name)

    # currently the partitioning expressions change the name of the column
    # so we need to alias it back to the original column name
    return transform_expr


def to_partition_representation(value: Any) -> Any:
    """Converts a partition value to the format expected by Iceberg metadata.

    Most transforms already do this, but the identity transforms preserve the original value type so we need to convert it.
    """
    if value is None:
        return None

    if isinstance(value, datetime.datetime):
        # Convert to microseconds since epoch
        return (value - datetime.datetime(1970, 1, 1)) // datetime.timedelta(microseconds=1)
    elif isinstance(value, datetime.date):
        # Convert to days since epoch
        return (value - datetime.date(1970, 1, 1)) // datetime.timedelta(days=1)
    elif isinstance(value, datetime.time):
        # Convert to microseconds since midnight
        return (value.hour * 60 * 60 + value.minute * 60 + value.second) * 1_000_000 + value.microsecond
    elif isinstance(value, uuid.UUID):
        return str(value)
    else:
        return value


def make_iceberg_data_file(
    file_path: str,
    size: int,
    metadata: pq.FileMetaData,
    partition_record: IcebergRecord,
    spec_id: int,
    schema: IcebergSchema,
    properties: dict[str, str],
) -> DataFile:
    import pyiceberg
    from packaging.version import parse
    from pyiceberg.io.pyarrow import (
        compute_statistics_plan,
        parquet_path_to_id_mapping,
    )
    from pyiceberg.manifest import DataFile, DataFileContent
    from pyiceberg.manifest import FileFormat as IcebergFileFormat

    kwargs = {
        "content": DataFileContent.DATA,
        "file_path": file_path,
        "file_format": IcebergFileFormat.PARQUET,
        "partition": partition_record,
        "file_size_in_bytes": size,
        # After this has been fixed:
        # https://github.com/apache/iceberg-python/issues/271
        # "sort_order_id": task.sort_order_id,
        "sort_order_id": None,
        # Just copy these from the table for now
        "spec_id": spec_id,
        "equality_ids": None,
        "key_metadata": None,
    }

    if parse(pyiceberg.__version__) >= parse("0.7.0"):
        from pyiceberg.io.pyarrow import data_file_statistics_from_parquet_metadata

        statistics = data_file_statistics_from_parquet_metadata(
            parquet_metadata=metadata,
            stats_columns=compute_statistics_plan(schema, properties),
            parquet_column_mapping=parquet_path_to_id_mapping(schema),
        )

        data_file = DataFile(
            **{
                **kwargs,
                **statistics.to_serialized_dict(),
            }
        )
    else:
        from pyiceberg.io.pyarrow import fill_parquet_file_metadata

        data_file = DataFile(**kwargs)

        fill_parquet_file_metadata(
            data_file=data_file,
            parquet_metadata=metadata,
            stats_columns=compute_statistics_plan(schema, properties),
            parquet_column_mapping=parquet_path_to_id_mapping(schema),
        )

    return data_file


class IcebergWriteVisitors:
    class FileVisitor:
        def __init__(self, parent: IcebergWriteVisitors, partition_record: IcebergRecord):
            self.parent = parent
            self.partition_record = partition_record

        def __call__(self, written_file: pads.WrittenFile) -> None:
            file_path = f"{self.parent.protocol}://{written_file.path}"
            data_file = make_iceberg_data_file(
                file_path,
                written_file.size,
                written_file.metadata,
                self.partition_record,
                self.parent.spec_id,
                self.parent.schema,
                self.parent.properties,
            )

            self.parent.data_files.append(data_file)

    def __init__(
        self,
        protocol: str,
        spec_id: int,
        schema: IcebergSchema,
        properties: IcebergTableProperties,
    ):
        self.data_files: list[DataFile] = []
        self.protocol = protocol
        self.spec_id = spec_id
        self.schema = schema
        self.properties = properties

    def visitor(self, partition_record: IcebergRecord) -> IcebergWriteVisitors.FileVisitor:
        return self.FileVisitor(self, partition_record)

    def to_metadata(self) -> MicroPartition:
        col_name = "data_file"
        if len(self.data_files) == 0:
            return MicroPartition.empty(_get_schema_from_dict({col_name: DataType.python()}))
        return MicroPartition.from_pydict({col_name: self.data_files})


def make_iceberg_record(partition_values: dict[str, Any] | None) -> IcebergRecord:
    from pyiceberg.typedef import Record as IcebergRecord

    if partition_values:
        iceberg_part_vals = {k: to_partition_representation(v) for k, v in partition_values.items()}
        return IcebergRecord(**iceberg_part_vals)
    else:
        return IcebergRecord()


def partitioned_table_to_iceberg_iter(
    partitioned: PartitionedTable, root_path: str, schema: pa.Schema
) -> Iterator[tuple[pa.Table, str, IcebergRecord]]:
    partition_values = partitioned.partition_values()

    if partition_values:
        partition_strings = partitioned.partition_values_str()
        assert partition_strings is not None

        for table, part_vals, part_strs in zip(
            partitioned.partitions(),
            partition_values.to_pylist(),
            partition_strings.to_pylist(),
        ):
            part_record = make_iceberg_record(part_vals)
            part_path = partition_strings_to_path(root_path, part_strs, partition_null_fallback="null")

            arrow_table = coerce_pyarrow_table_to_schema(table.to_arrow(), schema)

            yield arrow_table, part_path, part_record
    else:
        arrow_table = coerce_pyarrow_table_to_schema(partitioned.table.to_arrow(), schema)

        yield arrow_table, root_path, make_iceberg_record(None)



from __future__ import annotations

import numpy as np
import pytest

import daft
from daft import DataType

MODES = ["L", "LA", "RGB", "RGBA"]
MODE_TO_NP_DTYPE = {
    "L": np.uint8,
    "LA": np.uint8,
    "RGB": np.uint8,
    "RGBA": np.uint8,
    "L16": np.uint16,
    "LA16": np.uint16,
    "RGB16": np.uint16,
    "RGBA16": np.uint16,
    "RGB32F": np.float32,
    "RGBA32F": np.float32,
}
MODE_TO_NUM_CHANNELS = {
    "L": 1,
    "LA": 2,
    "RGB": 3,
    "RGBA": 4,
    "L16": 1,
    "LA16": 2,
    "RGB16": 3,
    "RGBA16": 4,
    "RGB32F": 3,
    "RGBA32F": 4,
}


@pytest.fixture(scope="session", params=MODES)
def fixed_shape_data_fixture(request):
    mode = request.param
    np_dtype = MODE_TO_NP_DTYPE[mode]
    num_channels = MODE_TO_NUM_CHANNELS[mode]

    height = 3
    width = 4
    dtype = daft.DataType.image(mode, height, width)
    shape = (height, width, num_channels)
    arr = np.arange(np.prod(shape)).reshape(shape).astype(np_dtype)
    arrs = [arr, arr, None]

    if mode in ("LA", "RGBA"):
        for arr in arrs:
            if arr is not None:
                arr[..., -1] = 255

    s = daft.Series.from_pylist(arrs, dtype=DataType.python())
    return s.cast(dtype)


@pytest.fixture(scope="session", params=MODES)
def mixed_shape_data_fixture(request):
    mode = request.param
    np_dtype = MODE_TO_NP_DTYPE[mode]
    num_channels = MODE_TO_NUM_CHANNELS[mode]

    dtype = daft.DataType.image(mode)
    shape1 = (2, 3, num_channels)
    shape2 = (3, 4, num_channels)
    arr1 = np.arange(np.prod(shape1)).reshape(shape1).astype(np_dtype)
    arr2 = np.arange(np.prod(shape1), np.prod(shape1) + np.prod(shape2)).reshape(shape2).astype(np_dtype)
    arrs = [arr1, arr2, None]

    if mode in ("LA", "RGBA"):
        for arr in arrs:
            if arr is not None:
                arr[..., -1] = 255

    s = daft.Series.from_pylist(arrs, dtype=DataType.python())
    return s.cast(dtype)



from __future__ import annotations

import daft
from daft.datatype import DataType
from daft.expressions import col
from daft.recordbatch import MicroPartition


def test_list_sort():
    table = MicroPartition.from_pydict(
        {
            "a": [[1, 3], [4, 6, 2], [3], [11, 6], None, [], [1, None, 3], None],
            "b": [True, False, False, True, False, False, False, True],
        }
    ).eval_expression_list([col("a").cast(DataType.list(DataType.int64())), col("b")])

    res = table.eval_expression_list(
        [
            col("a").list.sort().alias("asc"),
            col("a").list.sort(True).alias("desc"),
            col("a").list.sort(col("b")).alias("mixed"),
        ]
    )

    assert res.to_pydict() == {
        "asc": [[1, 3], [2, 4, 6], [3], [6, 11], None, [], [1, 3, None], None],
        "desc": [[3, 1], [6, 4, 2], [3], [11, 6], None, [], [None, 3, 1], None],
        "mixed": [[3, 1], [2, 4, 6], [3], [11, 6], None, [], [1, 3, None], None],
    }


def test_list_sort_fixed_size():
    table = MicroPartition.from_pydict(
        {
            "a": [[1, 3], [6, 2], [3, 3], [11, 6], None, [2, None], None],
            "b": [True, False, False, True, False, False, True],
        }
    ).eval_expression_list([col("a").cast(DataType.fixed_size_list(DataType.int64(), 2)), col("b")])

    res = table.eval_expression_list(
        [
            col("a").list.sort().alias("asc"),
            col("a").list.sort(True).alias("desc"),
            col("a").list.sort(col("b")).alias("mixed"),
        ]
    )

    assert res.to_pydict() == {
        "asc": [[1, 3], [2, 6], [3, 3], [6, 11], None, [2, None], None],
        "desc": [[3, 1], [6, 2], [3, 3], [11, 6], None, [None, 2], None],
        "mixed": [[3, 1], [2, 6], [3, 3], [11, 6], None, [2, None], None],
    }


def test_list_sort_with_groupby():
    df = daft.from_pydict({"group_col": [1, 1, 1, 2, 2, 2], "id_col": ["c", "a", "b", "e", "d", "a"]})

    # Group by group_col, aggregate id_col as a list.
    grouped_df = df.groupby("group_col").agg(daft.col("id_col").agg_list().alias("ids_col"))

    # Sort the list.
    result = grouped_df.select(col("group_col"), col("ids_col").list.sort()).sort("group_col", desc=False)
    result_dict = result.to_pydict()
    expected = {"group_col": [1, 2], "ids_col": [["a", "b", "c"], ["a", "d", "e"]]}
    assert result_dict == expected

    # Cast to fixed size list and sort.
    result = grouped_df.select(
        col("group_col"), col("ids_col").cast(DataType.fixed_size_list(DataType.string(), 3)).list.sort()
    ).sort("group_col", desc=False)
    result_dict = result.to_pydict()
    expected = {"group_col": [1, 2], "ids_col": [["a", "b", "c"], ["a", "d", "e"]]}
    assert result_dict == expected


# Reproduce issue #4862.
def test_list_sort_groupby_larger_than_morsel_size():
    # Test that reproduces a bug with large datasets and groupby operations that exceed the default morsel size.This would previously fail with offset mismatch errors.
    import itertools

    morsel_size = daft.context.get_context().daft_execution_config.default_morsel_size
    num_groups = int(morsel_size + 10)
    expected_records_per_group = 1
    num_records = num_groups * expected_records_per_group

    group_ids = list(itertools.chain(*[expected_records_per_group * [f"{i}"] for i in range(num_groups)]))
    record_ids = [f"r-{i}" for i in range(num_records)]

    result = (
        daft.from_pydict(
            {
                "group_id": group_ids,
                "record_id": record_ids,
            }
        )
        .groupby("group_id")
        .agg(daft.col("record_id").agg_list().alias("record_ids"))
        .with_column(
            "record_ids_key",
            daft.col("record_ids").list.sort(),
        )
        .select(
            "group_id",
            "record_ids_key",
        )
    )

    result_dict = result.to_pydict()

    assert (
        len(result_dict["record_ids_key"]) == num_groups
    ), f"Expected {num_groups} groups, got {len(result_dict['record_ids_key'])}"

    # Each group should have exactly expected_records_per_group records: Group i gets records r-i*expected_records to r-(i+1)*expected_records-1
    for group_id, record_list in zip(result_dict["group_id"], result_dict["record_ids_key"]):
        assert (
            len(record_list) == expected_records_per_group
        ), f"Group {group_id} should have {expected_records_per_group} records, got {len(record_list)}"

        start_record = int(group_id) * expected_records_per_group
        expected_sorted = sorted([f"r-{start_record + i}" for i in range(expected_records_per_group)])
        assert (
            record_list == expected_sorted
        ), f"Group {group_id} should be sorted: {record_list[:]}... vs {expected_sorted[:]}..."
